package de.mmadler.filecopy.gui;
	
import java.awt.AWTException;
import java.awt.MenuItem;
import java.awt.PopupMenu;
import java.awt.SystemTray;
import java.awt.Toolkit;
import java.awt.TrayIcon;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Locale;
import java.util.ResourceBundle;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.mmadler.filecopy.utils.Configuration;
import de.mmadler.filecopy.utils.ConfigurationReader;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.event.EventHandler;
import javafx.fxml.FXMLLoader;
import javafx.fxml.Initializable;
import javafx.fxml.JavaFXBuilderFactory;
import javafx.stage.Modality;
import javafx.stage.Stage;
import javafx.stage.StageStyle;
import javafx.stage.WindowEvent;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.image.Image;
import javafx.scene.paint.Color;


public class FileCopy extends Application {
	public enum DialogType { OverwriteDialog, ErrorMessage };
	
	private static String _version;
	public static String getVersion() { return _version; }
	
	private Stage _mainWindow;
	private FileCopyController _mainController;
	private Stage _dialog;
	private IDialogController _dialogController;
	
	private static Configuration _config;
	public static Configuration getConfig() { return _config; }
	public static void setConfig(Configuration config) { _config = config; }
	public static ResourceBundle _locale;
	public static ResourceBundle getLocale() { return _locale; }
	
	private static Logger logger = null;
	
	private static final double WINDOW_WIDTH = 550.0;
	private static final double WINDOW_HEIGHT = 390.0;
	
	private static String _argDirectory;
	
	public static void main(String[] args) {
		try {
			InputStream versionFile = FileCopy.class.getClassLoader().getResourceAsStream("VERSION");
			InputStreamReader versionReader = new InputStreamReader(versionFile);
			BufferedReader versionBufferedReader = new BufferedReader(versionReader);
			_version = versionBufferedReader.readLine();
			versionBufferedReader.close();
			versionReader.close();
			versionFile.close();
		} catch (IOException e) {
			if(logger!=null) logger.error("Error reading version file: ", e);
			_version = "0.0.0";
		}

		if(args.length>0)
			_argDirectory = args[0];
		
		_config = ConfigurationReader.readConfiguration();
		
		if(_config.writeLog())
			logger = LoggerFactory.getLogger(FileCopy.class);
		
		_locale = ResourceBundle.getBundle("de.mmadler.filecopy.gui.locale.FileCopy", Locale.getDefault());
		
		launch(args);
	}

	@Override
	public void start(final Stage stage) {
		_mainWindow = stage;
		try {
			_mainWindow.setTitle("FileCopy (v"+_version+")");
			_mainWindow.setResizable(false);
			_mainWindow.getIcons().add(new Image("images/icons/filecopy_16.png"));
			_mainWindow.getIcons().add(new Image("images/icons/filecopy_32.png"));
			_mainWindow.getIcons().add(new Image("images/icons/filecopy_24.png"));
			_mainWindow.getIcons().add(new Image("images/icons/filecopy_64.png"));
			_mainWindow.setMinWidth(WINDOW_WIDTH);
			_mainWindow.setMinHeight(WINDOW_HEIGHT);
			initFileCopy();
			if(SystemTray.isSupported() && _config.showTrayIcon()) {
				Platform.setImplicitExit(false);
				setTrayIcon(stage);
			}
			_mainWindow.show();
			
			_dialog = new Stage(StageStyle.TRANSPARENT);
			_dialog.initModality(Modality.WINDOW_MODAL);
			_dialog.initOwner(_mainWindow);
			initOverwriteDialog();
		} catch(Exception e) {
			if(logger!=null) logger.error("Exception starting EWBEntityTransferService: ", e);
		}
	}
	
	private void initFileCopy() {
		try {
			_mainController = (FileCopyController) getSceneContent(_mainWindow, "FileCopy.fxml");
			_mainController.setMainGui(this);
			if(_argDirectory != null)
				((FileCopyController) _mainController).setTargetDirectory(_argDirectory);
		} catch (Exception e) {
			if(logger!=null) logger.error("Exception opening file copy window: ", e);
		}	
	}
	
	private void initOverwriteDialog() {
		try {
			FXMLLoader loader = new FXMLLoader();
			loader.setBuilderFactory(new JavaFXBuilderFactory());
			loader.setLocation(getClass().getResource("OverwriteDialog.fxml"));
			loader.setResources(_locale);
			
			Parent page = (Parent) loader.load();

			Scene scene = new Scene(page, OverwriteController.dialogWidth, OverwriteController.dialogHeight);
			scene.setFill(Color.TRANSPARENT);
			_dialog.setScene(scene);
			_dialog.sizeToScene();
			_dialog.getScene().getStylesheets().add("FileCopy.css");
			_dialog.getScene().getStylesheets().add("Dialog.css");

			_dialogController = (OverwriteController) loader.getController();
			_dialogController.setDialog(_dialog);
		} catch (Exception e) {
			if(logger!=null) logger.error("Exception opening overwrite dialog: ", e);
		}	
	}
	
	private void initErrorMessageDialog() {
		try {
			FXMLLoader loader = new FXMLLoader();
			loader.setBuilderFactory(new JavaFXBuilderFactory());
			loader.setLocation(getClass().getResource("ErrorMessageDialog.fxml"));
			loader.setResources(_locale);
			
			Parent page = (Parent) loader.load();

			Scene scene = new Scene(page, ErrorMessageController.dialogWidth, ErrorMessageController.dialogHeight);
			scene.setFill(Color.TRANSPARENT);
			_dialog.setScene(scene);
			_dialog.sizeToScene();
			_dialog.getScene().getStylesheets().add("FileCopy.css");
			_dialog.getScene().getStylesheets().add("Dialog.css");

			_dialogController = loader.getController();
			_dialogController.setDialog(_dialog);
		} catch (Exception e) {
			if(logger!=null) logger.error("Exception opening overwrite dialog: ", e);
		}	
	}
	
	private Initializable getSceneContent(Stage stage, String fxml) throws IOException {
		FXMLLoader loader = new FXMLLoader();
		loader.setBuilderFactory(new JavaFXBuilderFactory());
		loader.setLocation(getClass().getResource(fxml));
		loader.setResources(_locale);
		
		Parent page = (Parent) loader.load();

		Scene scene = new Scene(page, WINDOW_WIDTH, WINDOW_HEIGHT);
		stage.setScene(scene);
		stage.sizeToScene();
		stage.getScene().getStylesheets().add("FileCopy.css");
		return (Initializable) loader.getController();
	}
	
	public Stage getMainWindow() {
		return _mainWindow;
	}

	public Stage getDialog(DialogType dialogType) {
		switch (dialogType) {
			case OverwriteDialog:
				initOverwriteDialog();
				break;
			case ErrorMessage:
				initErrorMessageDialog();
				break;
		}
		return _dialog;
	}
	public IDialogController getDialogController() {
		return _dialogController;
	}


	private void setTrayIcon(final Stage window) {
		SystemTray sTray = null;
        sTray = SystemTray.getSystemTray();
        java.awt.Image image = Toolkit.getDefaultToolkit().createImage(getClass().getResource("/images/icons/filecopy_16.png"));

        ActionListener listenerShow = new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                Platform.runLater(new Runnable() {
 
                    @Override
                    public void run() {
                        window.show();
                    }
                });
            }
        };
 
        ActionListener listenerClose = new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                System.exit(0);
            }
        };
 
        window.setOnCloseRequest(new EventHandler<WindowEvent>() {
            @Override
            public void handle(WindowEvent arg0) {
                window.hide();
            }
        });
        
        PopupMenu popup = new PopupMenu();
        MenuItem title = new MenuItem("FileCopy (v"+_version+")");
        title.setEnabled(false);
        MenuItem showItem = new MenuItem(_locale.getString("sysTray.open"));
        MenuItem exitItem = new MenuItem(_locale.getString("sysTray.exit"));
        showItem.addActionListener(listenerShow);
        exitItem.addActionListener(listenerClose);
        popup.add(title);
        popup.addSeparator();
        popup.add(showItem);
        popup.add(exitItem);
 
        TrayIcon icon = new TrayIcon(image, "FileCopy", popup);
 
        try {
            sTray.add(icon);
        }
        catch (AWTException e) {
            System.err.println(e);
        }
	}
}
