package de.mmadler.filecopy.gui;

import java.io.File;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.ResourceBundle;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.mmadler.filecopy.gui.FileCopy.DialogType;
import de.mmadler.filecopy.tasks.CopyFile;
import de.mmadler.filecopy.utils.Configuration;
import de.mmadler.filecopy.utils.ConfigurationReader;
import de.mmadler.filecopy.utils.OnErrorMode;
import de.mmadler.filecopy.utils.OverwriteMode;
import javafx.concurrent.WorkerStateEvent;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.geometry.Point2D;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Hyperlink;
import javafx.scene.control.Label;
import javafx.scene.control.ListView;
import javafx.scene.control.ProgressBar;
import javafx.scene.control.TextField;
import javafx.scene.control.Tooltip;
import javafx.scene.effect.BoxBlur;
import javafx.scene.input.DragEvent;
import javafx.scene.input.Dragboard;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseEvent;
import javafx.scene.input.TransferMode;
import javafx.scene.layout.AnchorPane;
import javafx.stage.DirectoryChooser;
import javafx.stage.Stage;

public class FileCopyController extends AnchorPane implements Initializable {
	@FXML
	private ComboBox<String> fxTargetDirectory;
	@FXML
	private Label fxFileCopyingLabel;
	@FXML
	private ProgressBar fxFileCopyingProgress;
	@FXML
	private Label fxFileCopyingProgressLabel;
	@FXML
	private Label fxProgressTime;
	@FXML
	private Label fxProgressSpeed;
	@FXML
	private ListView<File> fxFileList;
	@FXML
	private Button fxClearList;
	@FXML
	private Button fxListMoveUp;
	@FXML
	private Button fxListMoveDown;
	@FXML
	private Button fxStartCopying;
	
	@FXML
	private TextField fxDefaultTargetDirectory;
	@FXML
	private TextField fxRememberDirectories;
	@FXML
	private ChoiceBox<OverwriteMode> fxOverwriteMode;
	@FXML
	private ChoiceBox<OnErrorMode> fxOnErrorMode;
	@FXML
	private CheckBox fxKeepTimeStamp;
	@FXML
	private CheckBox fxTrayIcon;
	@FXML
	private CheckBox fxWriteLog;
	
	@FXML
	private Label fxAboutVersion;
	
	private CopyFile _task;
	private Thread _taskThread;
	
	private List<File> _skippedFiles;

	private FileCopy _mainAppl;
	private ResourceBundle _locale;
	
	private static Logger logger = null;
	
	/* ********************************************************************************** */
	/* *** Initialize                                                                 *** */
	/* ********************************************************************************** */
	
	@Override
	public void initialize(URL arg0, ResourceBundle arg1) {
		_locale = arg1;
		
		_task = null;
		_taskThread = null;
		_skippedFiles = new ArrayList<File>();
		
		if(FileCopy.getConfig().writeLog())
			logger = LoggerFactory.getLogger(CopyFile.class);
		
		for(String directory : FileCopy.getConfig().getTargetDirectories())
			fxTargetDirectory.getItems().add(directory);
		if(FileCopy.getConfig().getDefaultDirectory()!=null)
			fxTargetDirectory.setValue(FileCopy.getConfig().getDefaultDirectory());

		fxFileList.setOnDragOver(new EventHandler<DragEvent>() {
			@Override
			public void handle(DragEvent event) {
				Dragboard db = event.getDragboard();
				if(db.hasFiles())
					event.acceptTransferModes(TransferMode.COPY);
				else
					event.consume();
			}
		});
		fxFileList.setOnDragDropped(new EventHandler<DragEvent>() {
			private void addDirectory(File dir) {
				File[] files = dir.listFiles();
				Arrays.sort(files);
				for(File file : files) {
					if(file.isDirectory())
						addDirectory(file);
					else
						fxFileList.getItems().add(file);
				}
			}
			@Override
			public void handle(DragEvent event) {
				Dragboard db = event.getDragboard();
				boolean success = false;
				if(db.hasFiles()) {
					success = true;
					for(File file : db.getFiles()) {
						if(file.isDirectory())
							addDirectory(file);
						else
							fxFileList.getItems().add(file);
					}
					event.setDropCompleted(success);
					event.consume();
				}
			}
		});
		fxFileList.setOnKeyPressed(new EventHandler<KeyEvent>() {
			@Override
			public void handle(KeyEvent event) {
				List<KeyCode> keyCodes = Arrays.asList(new KeyCode[] { KeyCode.DELETE, KeyCode.BACK_SPACE });
				if(keyCodes.contains(event.getCode()) && fxFileList.getSelectionModel().getSelectedItem()!=null) {
					fxFileList.getItems().remove(fxFileList.getSelectionModel().getSelectedItem());
					event.consume();
				}
			}
		});
		
		fxDefaultTargetDirectory.setText(FileCopy.getConfig().getDefaultDirectory());
		fxRememberDirectories.setText(String.valueOf(FileCopy.getConfig().getRememberDirectories()));
		fxRememberDirectories.setTooltip(new Tooltip(_locale.getString("settings.rememberDorectories.tooltip")));
		fxOverwriteMode.getItems().addAll(Arrays.asList(OverwriteMode.values()));
		fxOverwriteMode.setValue(FileCopy.getConfig().getOverWriteMode());
		fxOnErrorMode.getItems().addAll(Arrays.asList(OnErrorMode.values()));
		fxOnErrorMode.setValue(FileCopy.getConfig().getOnErrorMode());
		fxKeepTimeStamp.setSelected(FileCopy.getConfig().keepTimestamp());
		fxTrayIcon.setSelected(FileCopy.getConfig().showTrayIcon());
		fxWriteLog.setSelected(FileCopy.getConfig().writeLog());
		fxAboutVersion.setText("("+FileCopy.getVersion()+")");
	}

	public void setMainGui(FileCopy mainAppl) {
		_mainAppl = mainAppl;
	}

	public FileCopy getMainGui() {
		return _mainAppl;
	}

	public void setTargetDirectory(String dir) {
		fxTargetDirectory.setValue(dir);
	}

	/* ********************************************************************************** */
	/* *** GUI                                                                        *** */
	/* ********************************************************************************** */
	
	/* Target Directory                                    ****************************** */
	@FXML
	protected void fxChooseTargetDirectory(ActionEvent event) {
		fxTargetDirectory.setValue(chooseTargetDirectory());
	}
	@FXML
	protected void fxChooseDefaultTargetDirectory(ActionEvent event) {
		fxDefaultTargetDirectory.setText(chooseTargetDirectory());
	}
	private String chooseTargetDirectory() {
		DirectoryChooser dc = new DirectoryChooser();
		dc.setTitle(_locale.getString("chooseTargetDirectory.title"));
		if(fxTargetDirectory.getValue() != null && !fxTargetDirectory.getValue().equals(""))
			dc.setInitialDirectory(new File (fxTargetDirectory.getValue()));
		
		File dir;
		if((dir = dc.showDialog(_mainAppl.getMainWindow()))!=null)
			return dir.getAbsolutePath();
		return "";
	}

	/* File List                                           ****************************** */
	@FXML
	protected void fxClearList() {
		fxFileList.getItems().removeAll(fxFileList.getItems());
	}
	@FXML
	protected void fxListMoveUp() {
		if(fxFileList.getSelectionModel().getSelectedItem()!=null) {
			File currentItem = fxFileList.getSelectionModel().getSelectedItem();
			int currentItemIndex = fxFileList.getItems().indexOf(currentItem);
			if(currentItemIndex>0) {
				try {
					File tmp = fxFileList.getItems().set(currentItemIndex, fxFileList.getItems().get(currentItemIndex-1));
					fxFileList.getItems().set(currentItemIndex-1, tmp);
					fxFileList.getSelectionModel().select(currentItemIndex-1);
				} catch (Exception e) {
				}
			}
		}
	}
	@FXML
	protected void fxListMoveDown() {
		if(fxFileList.getSelectionModel().getSelectedItem()!=null) {
			File currentItem = fxFileList.getSelectionModel().getSelectedItem();
			int currentItemIndex = fxFileList.getItems().indexOf(currentItem);
			if(currentItemIndex<fxFileList.getItems().size()) {
				try {
					File tmp = fxFileList.getItems().set(currentItemIndex, fxFileList.getItems().get(currentItemIndex+1));
					fxFileList.getItems().set(currentItemIndex+1, tmp);
					fxFileList.getSelectionModel().select(currentItemIndex+1);
				} catch (Exception e) {
				}
			}
		}		
	}
	
	@FXML
	protected void fxShowTooltip(MouseEvent e) {
		Button b = (Button)e.getSource();
		Point2D p = b.localToScene(0.0, 0.0);
		((Button)e.getSource()).getTooltip().show(b, 
				p.getX() + b.getScene().getX() + b.getScene().getWindow().getX() + 10.0,
		        p.getY() + b.getScene().getY() + b.getScene().getWindow().getY() + 18.0);
	}
	@FXML
	protected void fxHideTooltip(MouseEvent e) {
		((Button)e.getSource()).getTooltip().hide();
	}

	/* Copying                                             ****************************** */
	@FXML
	protected void fxStartCopying() {
		File targetDir = getTargetDirectory(fxTargetDirectory.getValue());
		if(targetDir==null || !targetDir.isDirectory()) { handleTargetIsNoDirectoryException(); return; }
		if(!targetDir.canWrite()) { handleTargetDirNotWritableException(); return; }

		_skippedFiles = new ArrayList<File>();
		fxTargetDirectory.setDisable(true);
		fxStartCopying.setDisable(true);
		storeTargetDirectory();
		
		copyFile(targetDir);
	}
	
	private void copyFile(final File targetDir) {
		boolean skipFile;
		File tmpSourceFile;
		do {
			skipFile = false;
			if(fxFileList.getItems().size()==0) {
				copyingFinished();
				return;
			}
			
			// Get first item from List
			tmpSourceFile = fxFileList.getItems().get(0);
			fxFileList.getItems().remove(0);
			
			// Init task
			_task = new CopyFile(tmpSourceFile, targetDir);
			if(_task.destinationExists() || !_task.destinationWritable()) {
				OverwriteMode om = FileCopy.getConfig().getOverWriteMode();
				if(om==OverwriteMode.Ask || !_task.destinationWritable()) 
					om = showOverwriteDialog(tmpSourceFile.getName(), _task.destinationWritable());
				switch(om) {
					case Overwrite:
						break;
					case Skip:
						skipFile = true;
					case Rename:
						_task.renameDestination();
						break;
					default:
						skipFile = true;
				}
			}
		} while(skipFile);
		
		final File sourceFile = tmpSourceFile;
		displayProgress(sourceFile.getAbsolutePath(), true, false);
		 
		//Init thread
		_taskThread = new Thread(_task);
		_task.setOnSucceeded(new EventHandler<WorkerStateEvent>() {
			@Override
			public void handle(WorkerStateEvent arg0) {
				displayProgress("", false, false);
				copyFile(targetDir);
			}
		});
		_task.setOnCancelled(new EventHandler<WorkerStateEvent>() {
			@Override
			public void handle(WorkerStateEvent arg0) {
				fxFileList.getItems().add(0, sourceFile);
				displayProgress(_locale.getString("cancelCopying.label"), false, true);
				copyingFinished();
			}
		});
		_task.setOnFailed(new EventHandler<WorkerStateEvent>() {
			@Override
			public void handle(WorkerStateEvent arg0) {
				switch(FileCopy.getConfig().getOnErrorMode()) {
					case Skip:
						if(logger!=null) logger.error(String.format(_locale.getString("errorMessage.copyingFailed"), sourceFile.getAbsolutePath()));
						_skippedFiles.add(sourceFile);
						displayProgress("", false, false);
						copyFile(targetDir);
						break;
					case Abort:
						fxFileList.getItems().add(0, sourceFile);
						handleCopyingFailedException(sourceFile.getAbsolutePath());
						copyingFinished();
						break;
				}
			}
		});
		_taskThread.start();
	}
	
	private void copyingFinished() {
		if(_skippedFiles.size()>0) {
			handleSkippedFiles();
			for(File skippedFile : _skippedFiles)
				fxFileList.getItems().add(skippedFile);
		}
		fxTargetDirectory.setDisable(false);
		fxStartCopying.setDisable(false);
	}
	
	@FXML
	protected void fxPauseCopying(ActionEvent event) {
		if(_task!=null && _task.isRunning()) {
			if(_task.isPaused()) {
				_task.resume();
				((Button)event.getTarget()).setText(_locale.getString("pauseCopying.button"));
			}
			else {
				_task.pause();
				((Button)event.getTarget()).setText(_locale.getString("resumeCopying.button"));
			}
		}
	}
	
	@FXML
	protected void fxCancelCopying() {
		if(_task!=null && _task.isRunning()) {
			_task.cancel(true);
		}
	}
	
	/* Configuration                                       ****************************** */
	@FXML
	protected void fxCommitConfig() {
		if(fxDefaultTargetDirectory==null) return;
		
		FileCopy.getConfig().setDefaultDirectory(fxDefaultTargetDirectory.getText());
		FileCopy.getConfig().setOverWriteMode(fxOverwriteMode.getValue());
		FileCopy.getConfig().setOnErrorMode(fxOnErrorMode.getValue());
		FileCopy.getConfig().setKeepTimestamp(fxKeepTimeStamp.isSelected());
		FileCopy.getConfig().setTrayIcon(fxTrayIcon.isSelected());
		FileCopy.getConfig().setWriteLog(fxWriteLog.isSelected());
	}
	
	@FXML
	protected void fxSaveConfig() {
		fxCommitConfig();
		ConfigurationReader.writeConfiguration(FileCopy.getConfig());
	}

	@FXML
	protected void fxRestoreConfig() {
		FileCopy.setConfig(ConfigurationReader.readConfiguration());

		fxDefaultTargetDirectory.setText(FileCopy.getConfig().getDefaultDirectory());
		fxOverwriteMode.setValue(FileCopy.getConfig().getOverWriteMode());
		fxOnErrorMode.setValue(FileCopy.getConfig().getOnErrorMode());
		fxKeepTimeStamp.setSelected(FileCopy.getConfig().getKeepTimestamp());
		fxTrayIcon.setSelected(FileCopy.getConfig().getTrayIcon());
		fxWriteLog.setSelected(FileCopy.getConfig().getWriteLog());
	}
	
	@FXML
	protected void fxOpenHyperlinkMailto(ActionEvent event) {
		try {
			java.awt.Desktop.getDesktop().browse(new URI("mailto:"+((Hyperlink)event.getSource()).getText()));
		} catch (Exception e) {
			//logger.error();
		}		
	}

	@FXML
	protected void fxOpenHyperlink(ActionEvent event) {
		try {
			java.awt.Desktop.getDesktop().browse(new URI(((Hyperlink)event.getSource()).getText()));
		} catch (Exception e) {
			if(logger!=null) logger.error(_locale.getString("errorMessate.cannotOpenHyperlink"), e);
		}		
	}
	
	/* ********************************************************************************** */
	/* *** Utilities                                                                  *** */
	/* ********************************************************************************** */
	
	private File getTargetDirectory(String dir) {
		if(dir==null || dir.equals("")) return null;
		
		File directory = new File(dir);
		if(directory.isDirectory())
			return directory;
		
		return null;
	}
	
	private void storeTargetDirectory() {
		String targetDirectory = fxTargetDirectory.getValue();
		fxTargetDirectory.getItems().remove(targetDirectory);
		fxTargetDirectory.getItems().add(0, targetDirectory);
		while(fxTargetDirectory.getItems().size() > FileCopy.getConfig().getRememberDirectories())
			fxTargetDirectory.getItems().remove(fxTargetDirectory.getItems().size()-1);
		fxTargetDirectory.setValue(targetDirectory);
		
		Configuration config = ConfigurationReader.readConfiguration();
		config.setTargetDirectories(new ArrayList<String>(fxTargetDirectory.getItems()));
		FileCopy.getConfig().setTargetDirectories(config.getTargetDirectories());
		ConfigurationReader.writeConfiguration(config);
	}
	
	private void displayProgress(String text, boolean copying, boolean reset) {
		if(text.equals(""))
			fxFileCopyingLabel.setText(_locale.getString("fileCopying.label.finished"));
		else if(!copying)
			fxFileCopyingLabel.setText(text);
		else
			fxFileCopyingLabel.setText(_locale.getString("fileCopying.label")+" "+text+" ...");

		if(copying) {
			fxFileCopyingProgress.progressProperty().bind(_task.progressProperty());
			fxFileCopyingProgressLabel.textProperty().bind(_task.percentCompleteProperty());
			fxProgressTime.textProperty().bind(_task.timeRunningProperty());
			fxProgressSpeed.textProperty().bind(_task.copyingSpeedProperty());
		}
		else {
			fxFileCopyingProgress.progressProperty().unbind();
			fxFileCopyingProgressLabel.textProperty().unbind();
			fxProgressTime.textProperty().unbind();
			fxProgressSpeed.textProperty().unbind();
			if(reset) {
				fxFileCopyingProgress.setProgress(0.0);
				fxFileCopyingProgressLabel.setText("");
				fxProgressTime.setText("");
				fxProgressSpeed.setText("");
			}
		}
	}

	private OverwriteMode showOverwriteDialog(String fileName, boolean destinationWritable) {
		Stage stage = _mainAppl.getMainWindow();
		Stage dialog = _mainAppl.getDialog(DialogType.OverwriteDialog);
		OverwriteController controller = (OverwriteController) _mainAppl.getDialogController();

		stage.getScene().getRoot().setEffect(new BoxBlur());
		dialog.setX(stage.getX() + (stage.getWidth()-OverwriteController.dialogWidth)/2);
		dialog.setY(stage.getY() + (stage.getHeight()-OverwriteController.dialogHeight)/2);
		controller.setDestinationWritable(destinationWritable);
		String label = String.format(_locale.getString("overwriteDialog.label"), fileName);
		if(!destinationWritable)
			label += "\n" + _locale.getString("overwriteDialog.label.notWritable");
		controller.setLabel(label);
		dialog.showAndWait();
		OverwriteMode om = controller.getOverwriteMode();
		fxOverwriteMode.setValue(FileCopy.getConfig().getOverWriteMode());
		stage.getScene().getRoot().setEffect(null);
		return om;
	}
	
	/* ********************************************************************************** */
	/* *** Exceptions                                                                 *** */
	/* ********************************************************************************** */
	
	private void showErrorDialog(String errorMessage) {
		Stage stage = _mainAppl.getMainWindow();
		Stage dialog = _mainAppl.getDialog(DialogType.ErrorMessage);
		IDialogController controller = _mainAppl.getDialogController();

		stage.getScene().getRoot().setEffect(new BoxBlur());
		dialog.setX(stage.getX() + (stage.getWidth()-ErrorMessageController.dialogWidth)/2);
		dialog.setY(stage.getY() + (stage.getHeight()-ErrorMessageController.dialogHeight)/2);
		controller.setLabel(errorMessage);
				
		dialog.showAndWait();
		stage.getScene().getRoot().setEffect(null);
	}
	
	private void handleSkippedFiles() {
		showErrorDialog(_locale.getString("errorMessage.skippedFiles"));
	}

	private void handleTargetIsNoDirectoryException() {
		showErrorDialog(_locale.getString("errorMessage.noTargetDir"));
	}
	
	private void handleTargetDirNotWritableException() {
		showErrorDialog(_locale.getString("errorMessage.targetDirNotWritable"));
	}
	
	private void handleCopyingFailedException(String fileName) {
		showErrorDialog(String.format(_locale.getString("errorMessage.copyingFailed"), fileName));
	}
}
